@@ -36,6 +36,10 @@
 #include <arpa/inet.h>
 #include <netdb.h>
 
+#if defined(AF_INET6) && defined(IN6_IS_ADDR_V4MAPPED)
+#define USE_IPV6
+#endif
+
 
 /* A few convenient defines. */
 
@@ -55,26 +59,26 @@
 typedef union {
     struct sockaddr sa;
     struct sockaddr_in sa_in;
-#ifdef HAVE_SOCKADDR_IN6
+#ifdef USE_IPV6
     struct sockaddr_in6 sa_in6;
-#endif /* HAVE_SOCKADDR_IN6 */
-#ifdef HAVE_SOCKADDR_STORAGE
     struct sockaddr_storage sa_stor;
-#endif /* HAVE_SOCKADDR_STORAGE */
+#endif /* USE_IPV6 */
     } httpd_sockaddr;
 
 /* A server. */
 typedef struct {
     char* binding_hostname;
     char* server_hostname;
-    int port;
+    unsigned short port;
     char* cgi_pattern;
     char* charset;
+    char* p3p;
+    int max_age;
     char* cwd;
     int listen4_fd, listen6_fd;
     int no_log;
     FILE* logfp;
-    int no_symlink;
+    int no_symlink_check;
     int vhost;
     int global_passwd;
     char* url_pattern;
@@ -88,7 +92,7 @@ typedef struct {
     httpd_server* hs;
     httpd_sockaddr client_addr;
     char* read_buf;
-    int read_size, read_idx, checked_idx;
+    size_t read_size, read_idx, checked_idx;
     int checked_state;
     int method;
     int status;
@@ -115,16 +119,16 @@ typedef struct {
     char* authorization;
     char* remoteuser;
     char* response;
-    int maxdecodedurl, maxorigfilename, maxexpnfilename, maxencodings,
+    size_t maxdecodedurl, maxorigfilename, maxexpnfilename, maxencodings,
 	maxpathinfo, maxquery, maxaccept, maxaccepte, maxreqhost, maxhostdir,
 	maxremoteuser, maxresponse;
 #ifdef TILDE_MAP_2
     char* altdir;
-    int maxaltdir;
+    size_t maxaltdir;
 #endif /* TILDE_MAP_2 */
-    int responselen;
+    size_t responselen;
     time_t if_modified_since, range_if;
-    off_t contentlength;
+    size_t contentlength;
     char* type;		/* not malloc()ed */
     char* hostname;	/* not malloc()ed */
     int mime_flag;
@@ -165,10 +169,11 @@ typedef struct {
 ** Return (httpd_server*) 0 on error.
 */
 extern httpd_server* httpd_initialize(
-    char* hostname, httpd_sockaddr* sa4P, httpd_sockaddr* sa6P, int port,
-    char* cgi_pattern, char* charset, char* cwd, int no_log, FILE* logfp,
-    int no_symlink, int vhost, int global_passwd, char* url_pattern,
-    char* local_pattern, int no_empty_referers );
+    char* hostname, httpd_sockaddr* sa4P, httpd_sockaddr* sa6P,
+    unsigned short port, char* cgi_pattern, char* charset, char* p3p,
+    int max_age, char* cwd, int no_log, FILE* logfp, int no_symlink_check,
+    int vhost, int global_passwd, char* url_pattern, char* local_pattern,
+    int no_empty_referers );
 
 /* Change the log file. */
 extern void httpd_set_logfp( httpd_server* hs, FILE* logfp );
@@ -252,16 +257,22 @@ extern char* httpd_err503form;
 extern char* httpd_method_str( int method );
 
 /* Reallocate a string. */
-extern void httpd_realloc_str( char** strP, int* maxsizeP, int size );
+extern void httpd_realloc_str( char** strP, size_t* maxsizeP, size_t size );
 
 /* Format a network socket to a string representation. */
 extern char* httpd_ntoa( httpd_sockaddr* saP );
 
 /* Set NDELAY mode on a socket. */
-void httpd_set_ndelay( int fd );
+extern void httpd_set_ndelay( int fd );
 
 /* Clear NDELAY mode on a socket. */
-void httpd_clear_ndelay( int fd );
+extern void httpd_clear_ndelay( int fd );
+
+/* Read the requested buffer completely, accounting for interruptions. */
+extern int httpd_read_fully( int fd, void* buf, size_t nbytes );
+
+/* Write the requested buffer completely, accounting for interruptions. */
+extern int httpd_write_fully( int fd, const void* buf, size_t nbytes );
 
 /* Generate debugging statistics syslog message. */
 extern void httpd_logstats( long secs );
