@@ -33,6 +33,7 @@
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/stat.h>
+#include <sys/wait.h>
 #include <sys/uio.h>
 
 #include <errno.h>
@@ -59,12 +60,17 @@
 #include "timers.h"
 #include "match.h"
 
+#ifndef SHUT_WR
+#define SHUT_WR 1
+#endif
+
 
 static char* argv0;
 static int debug;
-static int port;
+static unsigned short port;
 static char* dir;
-static int do_chroot, no_log, no_symlink, do_vhost, do_global_passwd;
+static char* data_dir;
+static int do_chroot, no_log, no_symlink_check, do_vhost, do_global_passwd;
 static char* cgi_pattern;
 static char* url_pattern;
 static int no_empty_referers;
@@ -75,11 +81,13 @@ static char* hostname;
 static char* pidfile;
 static char* user;
 static char* charset;
+static char* p3p;
+static int max_age;
 
 
 typedef struct {
     char* pattern;
-    long limit;
+    long max_limit, min_limit;
     long rate;
     off_t bytes_since_avg;
     int num_sending;
@@ -95,16 +103,14 @@ typedef struct {
     httpd_conn* hc;
     int tnums[MAXTHROTTLENUMS];         /* throttle indexes */
     int numtnums;
-    long limit;
-    time_t started_at;
-    Timer* idle_read_timer;
-    Timer* idle_send_timer;
+    long max_limit, min_limit;
+    time_t started_at, active_at;
     Timer* wakeup_timer;
     Timer* linger_timer;
     long wouldblock_delay;
     off_t bytes;
-    off_t bytes_sent;
     off_t bytes_to_send;
+    off_t bytes_sent;
     } connecttab;
 static connecttab* connects;
 static int numconnects, maxconnects;
@@ -121,10 +127,12 @@ static int httpd_conn_count;
 static httpd_server* hs = (httpd_server*) 0;
 int terminate = 0;
 time_t start_time, stats_time;
-long stats_connections, stats_bytes;
+long stats_connections;
+off_t stats_bytes;
 int stats_simultaneous;
 
 static int got_usr1;
+static int watchdog_flag;
 
 
 /* Forwards. */
@@ -144,10 +152,10 @@ static void handle_linger( connecttab* c, struct timeval* tvP );
 static int check_throttles( connecttab* c );
 static void clear_throttles( connecttab* c, struct timeval* tvP );
 static void update_throttles( ClientData client_data, struct timeval* nowP );
+static void finish_connection( connecttab* c, struct timeval* tvP );
 static void clear_connection( connecttab* c, struct timeval* tvP );
 static void really_clear_connection( connecttab* c, struct timeval* tvP );
-static void idle_read_connection( ClientData client_data, struct timeval* nowP );
-static void idle_send_connection( ClientData client_data, struct timeval* nowP );
+static void idle( ClientData client_data, struct timeval* nowP );
 static void wakeup_connection( ClientData client_data, struct timeval* nowP );
 static void linger_clear_connection( ClientData client_data, struct timeval* nowP );
 static void occasional( ClientData client_data, struct timeval* nowP );
@@ -158,6 +166,7 @@ static void logstats( struct timeval* nowP );
 static void thttpd_logstats( long secs );
 
 
+/* SIGTERM and SIGINT say to exit immediately. */
 static void
 handle_term( int sig )
     {
@@ -168,16 +177,62 @@ handle_term( int sig )
     }
 
 
+/* SIGCHLD - a chile process exitted, so we need to reap the zombie */
+static void
+handle_chld( int sig )
+    {
+    pid_t pid;
+    int status;
+
+#ifndef HAVE_SIGSET
+    (void) signal( SIGCHLD, handle_chld );	/* set up handler again */
+#endif /* ! HAVE_SIGSET */
+
+    /* Reap defunct children until there aren't any more. */
+    for (;;)
+	{
+#ifdef HAVE_WAITPID
+	pid = waitpid( (pid_t) -1, &status, WNOHANG );
+#else /* HAVE_WAITPID */
+	pid = wait3( &status, WNOHANG, (struct rusage*) 0 );
+#endif /* HAVE_WAITPID */
+	if ( (int) pid == 0 )		/* none left */
+	    break;
+	if ( (int) pid < 0 )
+	    {
+	    if ( errno == EINTR )	/* because of ptrace */
+		continue;
+	    /* ECHILD shouldn't happen with the WNOHANG option,
+	    ** but with some kernels it does anyway.  Ignore it.
+	    */
+	    if ( errno != ECHILD )
+		syslog( LOG_ERR, "child wait - %m" );
+	    break;
+	    }
+	}
+    }
+
+
+/* SIGHUP says to re-open the log file.  Generally this doesn't work,
+** due to running chrooted.  What you have to do instead is send a SIGUSR1
+** to shut down cleanly, and then restart thttpd.  But the HUP code is here
+** anyway in case you absolutely have to try it.
+*/
 static void
 handle_hup( int sig )
     {
     FILE* logfp;
 
-    if ( no_log )
+#ifndef HAVE_SIGSET
+    (void) signal( SIGHUP, handle_hup );	/* set up handler again */
+#endif /* ! HAVE_SIGSET */
+
+    got_usr1 = 0;
+    if ( no_log || hs == (httpd_server*) 0 )
 	return;
 
     /* Re-open the log file. */
-    if ( logfile != (char*) 0 )
+    if ( logfile != (char*) 0 && strcmp( logfile, "-" ) != 0 )
 	{
 	logfp = fopen( logfile, "a" );
 	if ( logfp == (FILE*) 0 )
@@ -191,20 +246,59 @@ handle_hup( int sig )
     }
 
 
+/* SIGUSR1 says to exit as soon as all current connections are done. */
 static void
 handle_usr1( int sig )
     {
+    if ( numconnects == 0 )
+	{
+	/* If there are no active connections we want to exit immediately
+	** here.  Not only is it faster, but without any connections the
+	** main loop won't wake up until the next new connection.
+	*/
+	shut_down();
+	syslog( LOG_NOTICE, "exiting" );
+	closelog();
+	exit( 0 );
+	}
+    /* Otherwise, just set a flag that we got the signal. */
     got_usr1 = 1;
     }
 
 
+/* SIGUSR2 says to generate the stats syslogs immediately. */
 static void
 handle_usr2( int sig )
     {
+#ifndef HAVE_SIGSET
+    (void) signal( SIGUSR2, handle_usr2 );	/* set up handler again */
+#endif /* ! HAVE_SIGSET */
+
     logstats( (struct timeval*) 0 );
     }
 
 
+/* SIGALRM is used as a watchdog. */
+static void
+handle_alrm( int sig )
+    {
+    /* If nothing has been happening */
+    if ( ! watchdog_flag )
+	{
+	/* Try changing dirs to someplace we can write. */
+	(void) chdir( "/tmp" );
+	/* Dump core. */
+	abort();
+	}
+    watchdog_flag = 0;
+
+#ifndef HAVE_SIGSET
+    (void) signal( SIGALRM, handle_alrm );	/* set up handler again */
+#endif /* ! HAVE_SIGSET */
+    (void) alarm( OCCASIONAL_TIME * 3 );	/* set up alarm again */
+    }
+
+
 int
 main( int argc, char** argv )
     {
@@ -235,14 +329,6 @@ main( int argc, char** argv )
     /* Handle command-line arguments. */
     parse_args( argc, argv );
 
-    /* Check port number. */
-    if ( port <= 0 )
-	{
-	syslog( LOG_CRIT, "illegal port number" );
-	(void) fprintf( stderr, "%s: illegal port number\n", argv0 );
-	exit( 1 );
-	}
-
     /* Read zone info now, in case we chroot(). */
     tzset();
 
@@ -270,6 +356,8 @@ main( int argc, char** argv )
 	    no_log = 1;
 	    logfp = (FILE*) 0;
 	    }
+	else if ( strcmp( logfile, "-" ) == 0 )
+	    logfp = stdout;
 	else
 	    {
 	    logfp = fopen( logfile, "a" );
@@ -333,7 +421,8 @@ main( int argc, char** argv )
 	** them to save file descriptors.
 	*/
 	(void) fclose( stdin );
-	(void) fclose( stdout );
+	if ( logfp != stdout )
+	    (void) fclose( stdout );
 	(void) fclose( stderr );
 
 	/* Daemonize - make ourselves a subprocess. */
@@ -401,14 +490,40 @@ main( int argc, char** argv )
 	    }
 	}
 
+    /* Switch directories again if requested. */
+    if ( data_dir != (char*) 0 )
+	{
+	if ( chdir( data_dir ) < 0 )
+	    {
+	    syslog( LOG_CRIT, "data_dir chdir - %m" );
+	    perror( "data_dir chdir" );
+	    exit( 1 );
+	    }
+	}
+
     /* Set up to catch signals. */
+#ifdef HAVE_SIGSET
+    (void) sigset( SIGTERM, handle_term );
+    (void) sigset( SIGINT, handle_term );
+    (void) sigset( SIGCHLD, handle_chld );
+    (void) sigset( SIGPIPE, SIG_IGN );          /* get EPIPE instead */
+    (void) sigset( SIGHUP, handle_hup );
+    (void) sigset( SIGUSR1, handle_usr1 );
+    (void) sigset( SIGUSR2, handle_usr2 );
+    (void) sigset( SIGALRM, handle_alrm );
+#else /* HAVE_SIGSET */
     (void) signal( SIGTERM, handle_term );
     (void) signal( SIGINT, handle_term );
+    (void) signal( SIGCHLD, handle_chld );
     (void) signal( SIGPIPE, SIG_IGN );          /* get EPIPE instead */
     (void) signal( SIGHUP, handle_hup );
-    got_usr1 = 0;
     (void) signal( SIGUSR1, handle_usr1 );
     (void) signal( SIGUSR2, handle_usr2 );
+    (void) signal( SIGALRM, handle_alrm );
+#endif /* HAVE_SIGSET */
+    got_usr1 = 0;
+    watchdog_flag = 0;
+    (void) alarm( OCCASIONAL_TIME * 3 );
 
     /* Initialize the timer package. */
     tmr_init();
@@ -419,8 +534,9 @@ main( int argc, char** argv )
     hs = httpd_initialize(
 	hostname,
 	gotv4 ? &sa4 : (httpd_sockaddr*) 0, gotv6 ? &sa6 : (httpd_sockaddr*) 0,
-	port, cgi_pattern, charset, cwd, no_log, logfp, no_symlink, do_vhost,
-	do_global_passwd, url_pattern, local_pattern, no_empty_referers );
+	port, cgi_pattern, charset, p3p, max_age, cwd, no_log, logfp,
+	no_symlink_check, do_vhost, do_global_passwd, url_pattern,
+	local_pattern, no_empty_referers );
     if ( hs == (httpd_server*) 0 )
 	exit( 1 );
 
@@ -430,6 +546,12 @@ main( int argc, char** argv )
 	syslog( LOG_CRIT, "tmr_create(occasional) failed" );
 	exit( 1 );
 	}
+    /* Set up the idle timer. */
+    if ( tmr_create( (struct timeval*) 0, idle, JunkClientData, 5 * 1000L, 1 ) == (Timer*) 0 )
+	{
+	syslog( LOG_CRIT, "tmr_create(idle) failed" );
+	exit( 1 );
+	}
     if ( numthrottles > 0 )
 	{
 	/* Set up the throttles timer. */
@@ -448,7 +570,8 @@ main( int argc, char** argv )
 	}
 #endif /* STATS_TIME */
     start_time = stats_time = time( (time_t*) 0 );
-    stats_connections = stats_bytes = 0L;
+    stats_connections = 0;
+    stats_bytes = 0;
     stats_simultaneous = 0;
 
     /* If we're root, try to become someone else. */
@@ -482,7 +605,7 @@ main( int argc, char** argv )
 	/* Check for unnecessary security exposure. */
 	if ( ! do_chroot )
 	    syslog(
-		LOG_CRIT,
+		LOG_WARNING,
 		"started as root without requesting chroot(), warning only" );
 	}
 
@@ -538,7 +661,7 @@ main( int argc, char** argv )
 	    }
 
 	/* Is it a new connection? */
-	if ( hs != (httpd_server*) 0 && hs->listen6_fd != -1 && 
+	if ( hs != (httpd_server*) 0 && hs->listen6_fd != -1 &&
 	     fdwatch_check_fd( hs->listen6_fd ) )
 	    {
 	    if ( handle_newconnect( &tv, hs->listen6_fd ) )
@@ -548,7 +671,7 @@ main( int argc, char** argv )
 		*/
 		continue;
 	    }
-	if ( hs != (httpd_server*) 0 && hs->listen4_fd != -1 && 
+	if ( hs != (httpd_server*) 0 && hs->listen4_fd != -1 &&
 	     fdwatch_check_fd( hs->listen4_fd ) )
 	    {
 	    if ( handle_newconnect( &tv, hs->listen4_fd ) )
@@ -566,15 +689,16 @@ main( int argc, char** argv )
 	    if ( c == (connecttab*) 0 )
 		continue;
 	    hc = c->hc;
-	    if ( c->conn_state == CNST_READING &&
-		 fdwatch_check_fd( hc->conn_fd ) )
-		handle_read( c, &tv );
-	    else if ( c->conn_state == CNST_SENDING &&
-		 fdwatch_check_fd( hc->conn_fd ) )
-		handle_send( c, &tv );
-	    else if ( c->conn_state == CNST_LINGERING &&
-		 fdwatch_check_fd( hc->conn_fd ) )
-		handle_linger( c, &tv );
+	    if ( ! fdwatch_check_fd( hc->conn_fd ) )
+		/* Something went wrong. */
+		clear_connection( c, &tv );
+	    else
+		switch ( c->conn_state )
+		    {
+		    case CNST_READING: handle_read( c, &tv ); break;
+		    case CNST_SENDING: handle_send( c, &tv ); break;
+		    case CNST_LINGERING: handle_linger( c, &tv ); break;
+		    }
 	    }
 	tmr_run( &tv );
 
@@ -583,8 +707,13 @@ main( int argc, char** argv )
 	    terminate = 1;
 	    if ( hs != (httpd_server*) 0 )
 		{
-		httpd_terminate( hs );
+		httpd_server* ths = hs;
 		hs = (httpd_server*) 0;
+		if ( ths->listen4_fd != -1 )
+		    fdwatch_del_fd( ths->listen4_fd );
+		if ( ths->listen6_fd != -1 )
+		    fdwatch_del_fd( ths->listen6_fd );
+		httpd_terminate( ths );
 		}
 	    }
 	}
@@ -605,13 +734,14 @@ parse_args( int argc, char** argv )
     debug = 0;
     port = DEFAULT_PORT;
     dir = (char*) 0;
+    data_dir = (char*) 0;
 #ifdef ALWAYS_CHROOT
     do_chroot = 1;
 #else /* ALWAYS_CHROOT */
     do_chroot = 0;
 #endif /* ALWAYS_CHROOT */
     no_log = 0;
-    no_symlink = do_chroot;
+    no_symlink_check = do_chroot;
 #ifdef ALWAYS_VHOST
     do_vhost = 1;
 #else /* ALWAYS_VHOST */
@@ -636,10 +766,17 @@ parse_args( int argc, char** argv )
     pidfile = (char*) 0;
     user = DEFAULT_USER;
     charset = DEFAULT_CHARSET;
+    p3p = "";
+    max_age = -1;
     argn = 1;
     while ( argn < argc && argv[argn][0] == '-' )
 	{
-	if ( strcmp( argv[argn], "-C" ) == 0 && argn + 1 < argc )
+	if ( strcmp( argv[argn], "-V" ) == 0 )
+	    {
+	    (void) printf( "%s\n", SERVER_SOFTWARE );
+	    exit( 0 );
+	    }
+	else if ( strcmp( argv[argn], "-C" ) == 0 && argn + 1 < argc )
 	    {
 	    ++argn;
 	    read_config( argv[argn] );
@@ -647,7 +784,7 @@ parse_args( int argc, char** argv )
 	else if ( strcmp( argv[argn], "-p" ) == 0 && argn + 1 < argc )
 	    {
 	    ++argn;
-	    port = atoi( argv[argn] );
+	    port = (unsigned short) atoi( argv[argn] );
 	    }
 	else if ( strcmp( argv[argn], "-d" ) == 0 && argn + 1 < argc )
 	    {
@@ -657,17 +794,22 @@ parse_args( int argc, char** argv )
 	else if ( strcmp( argv[argn], "-r" ) == 0 )
 	    {
 	    do_chroot = 1;
-	    no_symlink = 1;
+	    no_symlink_check = 1;
 	    }
 	else if ( strcmp( argv[argn], "-nor" ) == 0 )
 	    {
 	    do_chroot = 0;
-	    no_symlink = 0;
+	    no_symlink_check = 0;
+	    }
+	else if ( strcmp( argv[argn], "-dd" ) == 0 && argn + 1 < argc )
+	    {
+	    ++argn;
+	    data_dir = argv[argn];
 	    }
 	else if ( strcmp( argv[argn], "-s" ) == 0 )
-	    no_symlink = 0;
+	    no_symlink_check = 0;
 	else if ( strcmp( argv[argn], "-nos" ) == 0 )
-	    no_symlink = 1;
+	    no_symlink_check = 1;
 	else if ( strcmp( argv[argn], "-u" ) == 0 && argn + 1 < argc )
 	    {
 	    ++argn;
@@ -711,10 +853,15 @@ parse_args( int argc, char** argv )
 	    ++argn;
 	    charset = argv[argn];
 	    }
-	else if ( strcmp( argv[argn], "-V" ) == 0 )
+	else if ( strcmp( argv[argn], "-P" ) == 0 && argn + 1 < argc )
 	    {
-	    (void) fprintf( stderr, "%s\n", SERVER_SOFTWARE );
-	    exit( 0 );
+	    ++argn;
+	    p3p = argv[argn];
+	    }
+	else if ( strcmp( argv[argn], "-M" ) == 0 && argn + 1 < argc )
+	    {
+	    ++argn;
+	    max_age = atoi( argv[argn] );
 	    }
 	else if ( strcmp( argv[argn], "-D" ) == 0 )
 	    debug = 1;
@@ -731,7 +878,7 @@ static void
 usage( void )
     {
     (void) fprintf( stderr,
-"usage:  %s [-C configfile] [-p port] [-d dir] [-r|-nor] [-v|-nov] [-g|-nog] [-u user] [-c cgipat] [-t throttles] [-h host] [-l logfile] [-i pidfile] [-T charset] [-V] [-D]\n",
+"usage:  %s [-C configfile] [-p port] [-d dir] [-r|-nor] [-dd data_dir] [-s|-nos] [-v|-nov] [-g|-nog] [-u user] [-c cgipat] [-t throttles] [-h host] [-l logfile] [-i pidfile] [-T charset] [-P P3P] [-M maxage] [-V] [-D]\n",
 	argv0 );
     exit( 1 );
     }
@@ -760,11 +907,13 @@ read_config( char* filename )
 	if ( ( cp = strchr( line, '#' ) ) != (char*) 0 )
 	    *cp = '\0';
 
+	/* Skip leading whitespace. */
+	cp = line;
+	cp += strspn( cp, " \t\n\r" );
+
 	/* Split line into words. */
-	for ( cp = line; *cp != '\0'; cp = cp2 )
+	while ( *cp != '\0' )
 	    {
-	    /* Skip leading whitespace. */
-	    cp += strspn( cp, " \t\n\r" );
 	    /* Find next whitespace. */
 	    cp2 = cp + strcspn( cp, " \t\n\r" );
 	    /* Insert EOS and advance next-word pointer. */
@@ -784,7 +933,7 @@ read_config( char* filename )
 	    else if ( strcasecmp( name, "port" ) == 0 )
 		{
 		value_required( name, value );
-		port = atoi( value );
+		port = (unsigned short) atoi( value );
 		}
 	    else if ( strcasecmp( name, "dir" ) == 0 )
 		{
@@ -795,33 +944,38 @@ read_config( char* filename )
 		{
 		no_value_required( name, value );
 		do_chroot = 1;
-		no_symlink = 1;
+		no_symlink_check = 1;
 		}
 	    else if ( strcasecmp( name, "nochroot" ) == 0 )
 		{
 		no_value_required( name, value );
 		do_chroot = 0;
-		no_symlink = 0;
+		no_symlink_check = 0;
+		}
+	    else if ( strcasecmp( name, "data_dir" ) == 0 )
+		{
+		value_required( name, value );
+		data_dir = e_strdup( value );
 		}
 	    else if ( strcasecmp( name, "symlink" ) == 0 )
 		{
 		no_value_required( name, value );
-		no_symlink = 0;
+		no_symlink_check = 0;
 		}
 	    else if ( strcasecmp( name, "nosymlink" ) == 0 )
 		{
 		no_value_required( name, value );
-		no_symlink = 1;
+		no_symlink_check = 1;
 		}
 	    else if ( strcasecmp( name, "symlinks" ) == 0 )
 		{
 		no_value_required( name, value );
-		no_symlink = 0;
+		no_symlink_check = 0;
 		}
 	    else if ( strcasecmp( name, "nosymlinks" ) == 0 )
 		{
 		no_value_required( name, value );
-		no_symlink = 1;
+		no_symlink_check = 1;
 		}
 	    else if ( strcasecmp( name, "user" ) == 0 )
 		{
@@ -893,12 +1047,26 @@ read_config( char* filename )
 		value_required( name, value );
 		charset = e_strdup( value );
 		}
+	    else if ( strcasecmp( name, "p3p" ) == 0 )
+		{
+		value_required( name, value );
+		p3p = e_strdup( value );
+		}
+	    else if ( strcasecmp( name, "max_age" ) == 0 )
+		{
+		value_required( name, value );
+		max_age = atoi( value );
+		}
 	    else
 		{
 		(void) fprintf(
 		    stderr, "%s: unknown config option '%s'\n", argv0, name );
 		exit( 1 );
 		}
+
+	    /* Advance to next word. */
+	    cp = cp2;
+	    cp += strspn( cp, " \t\n\r" );
 	    }
 	}
 
@@ -950,22 +1118,22 @@ e_strdup( char* oldstr )
 static void
 lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockaddr* sa6P, size_t sa6_len, int* gotv6P )
     {
-#if defined(HAVE_GETADDRINFO) && defined(HAVE_GAI_STRERROR)
+#ifdef USE_IPV6
 
     struct addrinfo hints;
+    char portstr[10];
+    int gaierr;
     struct addrinfo* ai;
     struct addrinfo* ai2;
-    struct addrinfo* aiv4;
     struct addrinfo* aiv6;
-    int gaierr;
-    char strport[10];
+    struct addrinfo* aiv4;
 
-    memset( &hints, 0, sizeof(hints) );
-    hints.ai_family = AF_UNSPEC;
+    (void) memset( &hints, 0, sizeof(hints) );
+    hints.ai_family = PF_UNSPEC;
     hints.ai_flags = AI_PASSIVE;
     hints.ai_socktype = SOCK_STREAM;
-    (void) snprintf( strport, sizeof(strport), "%d", port );
-    if ( (gaierr = getaddrinfo( hostname, strport, &hints, &ai )) != 0 )
+    (void) snprintf( portstr, sizeof(portstr), "%d", (int) port );
+    if ( (gaierr = getaddrinfo( hostname, portstr, &hints, &ai )) != 0 )
 	{
 	syslog(
 	    LOG_CRIT, "getaddrinfo %.80s - %.80s",
@@ -976,72 +1144,67 @@ lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockad
 	exit( 1 );
 	}
 
-    /* Find the first IPv4 and IPv6 entries. */
-    aiv4 = (struct addrinfo*) 0;
+    /* Find the first IPv6 and IPv4 entries. */
     aiv6 = (struct addrinfo*) 0;
+    aiv4 = (struct addrinfo*) 0;
     for ( ai2 = ai; ai2 != (struct addrinfo*) 0; ai2 = ai2->ai_next )
 	{
 	switch ( ai2->ai_family )
 	    {
-	    case AF_INET:
-	    if ( aiv4 == (struct addrinfo*) 0 )
-		aiv4 = ai2;
-	    break;
-#if defined(AF_INET6) && defined(HAVE_SOCKADDR_IN6)
 	    case AF_INET6:
 	    if ( aiv6 == (struct addrinfo*) 0 )
 		aiv6 = ai2;
 	    break;
-#endif /* AF_INET6 && HAVE_SOCKADDR_IN6 */
+	    case AF_INET:
+	    if ( aiv4 == (struct addrinfo*) 0 )
+		aiv4 = ai2;
+	    break;
 	    }
 	}
 
-    if ( aiv4 == (struct addrinfo*) 0 )
-	*gotv4P = 0; 
-    else
-	{
-	if ( sa4_len < aiv4->ai_addrlen )
-	    {
-	    syslog(
-		LOG_CRIT, "%.80s - sockaddr too small (%d < %d)",
-		hostname, sa4_len, aiv4->ai_addrlen );
-	    exit( 1 );
-	    }
-	memset( sa4P, 0, sa4_len );
-	memcpy( sa4P, aiv4->ai_addr, aiv4->ai_addrlen );
-	*gotv4P = 1;
-	}
     if ( aiv6 == (struct addrinfo*) 0 )
-	*gotv6P = 0; 
+	*gotv6P = 0;
     else
 	{
 	if ( sa6_len < aiv6->ai_addrlen )
 	    {
 	    syslog(
-		LOG_CRIT, "%.80s - sockaddr too small (%d < %d)",
-		hostname, sa6_len, aiv6->ai_addrlen );
+		LOG_CRIT, "%.80s - sockaddr too small (%lu < %lu)",
+		hostname, (unsigned long) sa6_len,
+		(unsigned long) aiv6->ai_addrlen );
 	    exit( 1 );
 	    }
-	memset( sa6P, 0, sa6_len );
-	memcpy( sa6P, aiv6->ai_addr, aiv6->ai_addrlen );
+	(void) memset( sa6P, 0, sa6_len );
+	(void) memmove( sa6P, aiv6->ai_addr, aiv6->ai_addrlen );
 	*gotv6P = 1;
 	}
 
+    if ( aiv4 == (struct addrinfo*) 0 )
+	*gotv4P = 0;
+    else
+	{
+	if ( sa4_len < aiv4->ai_addrlen )
+	    {
+	    syslog(
+		LOG_CRIT, "%.80s - sockaddr too small (%lu < %lu)",
+		hostname, (unsigned long) sa4_len,
+		(unsigned long) aiv4->ai_addrlen );
+	    exit( 1 );
+	    }
+	(void) memset( sa4P, 0, sa4_len );
+	(void) memmove( sa4P, aiv4->ai_addr, aiv4->ai_addrlen );
+	*gotv4P = 1;
+	}
+
     freeaddrinfo( ai );
 
-#else /* HAVE_GETADDRINFO && HAVE_GAI_STRERROR */
+#else /* USE_IPV6 */
 
     struct hostent* he;
 
     *gotv6P = 0;
 
-    memset( sa4P, 0, sa4_len );
-#ifdef notdef
-    /* We don't really need to set sa_len. */
-#ifdef HAVE_SA_LEN
-    sa4P->sa_len = sa4_len;
-#endif /* HAVE_SA_LEN */
-#endif
+    (void) memset( sa4P, 0, sa4_len );
     sa4P->sa.sa_family = AF_INET;
     if ( hostname == (char*) 0 )
 	sa4P->sa_in.sin_addr.s_addr = htonl( INADDR_ANY );
@@ -1075,14 +1238,14 @@ lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockad
 		    argv0, hostname );
 		exit( 1 );
 		}
-	    (void) memcpy(
+	    (void) memmove(
 		&sa4P->sa_in.sin_addr.s_addr, he->h_addr, he->h_length );
 	    }
 	}
     sa4P->sa_in.sin_port = htons( port );
     *gotv4P = 1;
 
-#endif /* HAVE_GETADDRINFO && HAVE_GAI_STRERROR */
+#endif /* USE_IPV6 */
     }
 
 
@@ -1094,7 +1257,7 @@ read_throttlefile( char* throttlefile )
     char* cp;
     int len;
     char pattern[5000];
-    long limit;
+    long max_limit, min_limit;
     struct timeval tv;
 
     fp = fopen( throttlefile, "r" );
@@ -1126,7 +1289,11 @@ read_throttlefile( char* throttlefile )
 	    continue;
 
 	/* Parse line. */
-	if ( sscanf( buf, " %4900[^ \t] %ld", pattern, &limit ) != 2 || limit <= 0 )
+	if ( sscanf( buf, " %4900[^ \t] %ld-%ld", pattern, &min_limit, &max_limit ) == 3 )
+	    {}
+	else if ( sscanf( buf, " %4900[^ \t] %ld", pattern, &max_limit ) == 2 )
+	    min_limit = 0;
+	else
 	    {
 	    syslog( LOG_CRIT,
 		"unparsable line in %.80s - %.80s", throttlefile, buf );
@@ -1166,16 +1333,9 @@ read_throttlefile( char* throttlefile )
 	    }
 
 	/* Add to table. */
-	throttles[numthrottles].pattern = strdup( pattern );
-	if ( throttles[numthrottles].pattern == (char*) 0 )
-	    {
-	    syslog( LOG_CRIT, "out of memory copying a throttle pattern" );
-	    (void) fprintf(
-		stderr, "%s: out of memory copying a throttle pattern\n",
-		argv0 );
-	    exit( 1 );
-	    }
-	throttles[numthrottles].limit = limit;
+	throttles[numthrottles].pattern = e_strdup( pattern );
+	throttles[numthrottles].max_limit = max_limit;
+	throttles[numthrottles].min_limit = min_limit;
 	throttles[numthrottles].rate = 0;
 	throttles[numthrottles].bytes_since_avg = 0;
 	throttles[numthrottles].num_sending = 0;
@@ -1210,6 +1370,10 @@ shut_down( void )
 	{
 	httpd_server* ths = hs;
 	hs = (httpd_server*) 0;
+	if ( ths->listen4_fd != -1 )
+	    fdwatch_del_fd( ths->listen4_fd );
+	if ( ths->listen6_fd != -1 )
+	    fdwatch_del_fd( ths->listen6_fd );
 	httpd_terminate( ths );
 	}
     mmc_destroy();
@@ -1239,7 +1403,7 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	    /* Out of connection slots.  Run the timers, then the
 	    ** existing connections, and maybe we'll free up a slot
 	    ** by the time we get back here.
-	    **/
+	    */
 	    syslog( LOG_WARNING, "too many connections!" );
 	    tmr_run( tvP );
 	    return 0;
@@ -1265,22 +1429,21 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	/* Get the connection. */
 	switch ( httpd_get_conn( hs, listen_fd, c->hc ) )
 	    {
+	    /* Some error happened.  Run the timers, then the
+	    ** existing connections.  Maybe the error will clear.
+	    */
 	    case GC_FAIL:
+	    tmr_run( tvP );
+	    return 0;
+
+	    /* No more connections to accept for now. */
 	    case GC_NO_MORE:
 	    return 1;
 	    }
 	c->conn_state = CNST_READING;
 	++numconnects;
 	client_data.p = c;
-	c->idle_read_timer = tmr_create(
-	    tvP, idle_read_connection, client_data, IDLE_READ_TIMELIMIT * 1000L,
-	    0 );
-	if ( c->idle_read_timer == (Timer*) 0 )
-	    {
-	    syslog( LOG_CRIT, "tmr_create(idle_read_connection) failed" );
-	    exit( 1 );
-	    }
-	c->idle_send_timer = (Timer*) 0;
+	c->active_at = tvP->tv_sec;
 	c->wakeup_timer = (Timer*) 0;
 	c->linger_timer = (Timer*) 0;
 	c->bytes_sent = 0;
@@ -1311,7 +1474,7 @@ handle_read( connecttab* c, struct timeval* tvP )
 	if ( hc->read_size > 5000 )
 	    {
 	    httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
-	    clear_connection( c, tvP );
+	    finish_connection( c, tvP );
 	    return;
 	    }
 	httpd_realloc_str(
@@ -1322,18 +1485,29 @@ handle_read( connecttab* c, struct timeval* tvP )
     sz = read(
 	hc->conn_fd, &(hc->read_buf[hc->read_idx]),
 	hc->read_size - hc->read_idx );
-    /* Ignore EWOULDBLOCK errors.  At first glance you would think that
-    ** connections returned by fdwatch as readable should never give an
-    ** EWOULDBLOCK; however, this apparently can happen if a packet gets
-    ** garbled.
-    */
-    if ( sz == 0 || ( sz < 0 && ( errno != EWOULDBLOCK ) ) )
+    if ( sz == 0 )
 	{
 	httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
+	return;
+	}
+    if ( sz < 0 )
+	{
+	/* Ignore EWOULDBLOCK errors.  At first glance you would think that
+	** connections returned by fdwatch as readable should never give an
+	** EWOULDBLOCK; however, this apparently can happen if a packet gets
+	** garbled.
+	*/
+	if ( errno != EWOULDBLOCK )
+	    {
+	    httpd_send_err(
+		hc, 400, httpd_err400title, "", httpd_err400form, "" );
+	    finish_connection( c, tvP );
+	    }
 	return;
 	}
     hc->read_idx += sz;
+    c->active_at = tvP->tv_sec;
 
     /* Do we have a complete request yet? */
     switch ( httpd_got_request( hc ) )
@@ -1342,14 +1516,14 @@ handle_read( connecttab* c, struct timeval* tvP )
 	return;
 	case GR_BAD_REQUEST:
 	httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
 	return;
 	}
 
     /* Yes.  Try parsing and resolving it. */
     if ( httpd_parse_request( hc ) < 0 )
 	{
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
 	return;
 	}
 
@@ -1358,7 +1532,7 @@ handle_read( connecttab* c, struct timeval* tvP )
 	{
 	httpd_send_err(
 	    hc, 503, httpd_err503title, "", httpd_err503form, hc->encodedurl );
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
 	return;
 	}
 
@@ -1366,7 +1540,7 @@ handle_read( connecttab* c, struct timeval* tvP )
     if ( httpd_start_request( hc, tvP ) < 0 )
 	{
 	/* Something went wrong.  Close down the connection. */
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
 	return;
 	}
 
@@ -1376,6 +1550,8 @@ handle_read( connecttab* c, struct timeval* tvP )
 	c->bytes_sent = hc->init_byte_loc;
 	c->bytes_to_send = hc->end_byte_loc + 1;
 	}
+    else if ( hc->bytes_to_send < 0 )
+	c->bytes_to_send = 0;
     else
 	c->bytes_to_send = hc->bytes_to_send;
 
@@ -1384,13 +1560,13 @@ handle_read( connecttab* c, struct timeval* tvP )
 	{
 	/* No file address means someone else is handling it. */
 	c->bytes_sent = hc->bytes_sent;
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
 	return;
 	}
     if ( c->bytes_sent >= c->bytes_to_send )
 	{
 	/* There's nothing to send. */
-	clear_connection( c, tvP );
+	finish_connection( c, tvP );
 	return;
 	}
 
@@ -1399,16 +1575,6 @@ handle_read( connecttab* c, struct timeval* tvP )
     c->started_at = tvP->tv_sec;
     c->wouldblock_delay = 0;
     client_data.p = c;
-    tmr_cancel( c->idle_read_timer );
-    c->idle_read_timer = (Timer*) 0;
-    c->idle_send_timer = tmr_create(
-	tvP, idle_send_connection, client_data, IDLE_SEND_TIMELIMIT * 1000L,
-	0 );
-    if ( c->idle_send_timer == (Timer*) 0 )
-	{
-	syslog( LOG_CRIT, "tmr_create(idle_send_connection) failed" );
-	exit( 1 );
-	}
 
     fdwatch_del_fd( hc->conn_fd );
     fdwatch_add_fd( hc->conn_fd, c, FDW_WRITE );
@@ -1429,7 +1595,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 	/* No, just write the file. */
 	sz = write(
 	    hc->conn_fd, &(hc->file_address[c->bytes_sent]),
-	    MIN( c->bytes_to_send - c->bytes_sent, c->limit ) );
+	    MIN( c->bytes_to_send - c->bytes_sent, c->max_limit ) );
 	}
     else
 	{
@@ -1441,7 +1607,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 	iv[0].iov_base = hc->response;
 	iv[0].iov_len = hc->responselen;
 	iv[1].iov_base = &(hc->file_address[c->bytes_sent]);
-	iv[1].iov_len = MIN( c->bytes_to_send - c->bytes_sent, c->limit );
+	iv[1].iov_len = MIN( c->bytes_to_send - c->bytes_sent, c->max_limit );
 	sz = writev( hc->conn_fd, iv, 2 );
 	}
 
@@ -1462,6 +1628,8 @@ handle_send( connecttab* c, struct timeval* tvP )
 	c->conn_state = CNST_PAUSING;
 	fdwatch_del_fd( hc->conn_fd );
 	client_data.p = c;
+	if ( c->wakeup_timer != (Timer*) 0 )
+	    syslog( LOG_ERR, "replacing non-null wakeup_timer!" );
 	c->wakeup_timer = tmr_create(
 	    tvP, wakeup_connection, client_data, c->wouldblock_delay, 0 );
 	if ( c->wakeup_timer == (Timer*) 0 )
@@ -1491,7 +1659,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 	}
 
     /* Ok, we wrote something. */
-    tmr_reset( tvP, c->idle_send_timer );
+    c->active_at = tvP->tv_sec;
     /* Was this a headers + file writev()? */
     if ( hc->responselen > 0 )
 	{
@@ -1500,7 +1668,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 	    {
 	    /* Yes; move the unwritten part to the front of the buffer. */
 	    int newlen = hc->responselen - sz;
-	    (void) memcpy( hc->response, &(hc->response[sz]), newlen );
+	    (void) memmove( hc->response, &(hc->response[sz]), newlen );
 	    hc->responselen = newlen;
 	    sz = 0;
 	    }
@@ -1518,8 +1686,8 @@ handle_send( connecttab* c, struct timeval* tvP )
     /* Are we done? */
     if ( c->bytes_sent >= c->bytes_to_send )
 	{
-	/* This conection is finished! */
-	clear_connection( c, tvP );
+	/* This connection is finished! */
+	finish_connection( c, tvP );
 	return;
 	}
 
@@ -1528,19 +1696,21 @@ handle_send( connecttab* c, struct timeval* tvP )
 	c->wouldblock_delay -= MIN_WOULDBLOCK_DELAY;
 
     /* If we're throttling, check if we're sending too fast. */
-    if ( c->limit != THROTTLE_NOLIMIT )
+    if ( c->max_limit != THROTTLE_NOLIMIT )
 	{
 	elapsed = tvP->tv_sec - c->started_at;
-	if ( elapsed == 0 || c->hc->bytes_sent / elapsed > c->limit )
+	if ( elapsed == 0 || c->hc->bytes_sent / elapsed > c->max_limit )
 	    {
 	    c->conn_state = CNST_PAUSING;
 	    fdwatch_del_fd( hc->conn_fd );
-	    /* When should we send the next c->limit bytes
+	    /* When should we send the next c->max_limit bytes
 	    ** to get back on schedule?  If less than a second
 	    ** (integer math rounding), use 1/8 second.
 	    */
-	    coast = ( c->hc->bytes_sent + c->limit ) / c->limit - elapsed;
+	    coast = ( c->hc->bytes_sent + c->max_limit ) / c->max_limit - elapsed;
 	    client_data.p = c;
+	    if ( c->wakeup_timer != (Timer*) 0 )
+		syslog( LOG_ERR, "replacing non-null wakeup_timer!" );
 	    c->wakeup_timer = tmr_create(
 		tvP, wakeup_connection, client_data,
 		coast ? ( coast * 1000L ) : 125L, 0 );
@@ -1551,6 +1721,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 		}
 	    }
 	}
+    /* (No check on min_limit here, that only controls connection startups.) */
     }
 
 
@@ -1573,15 +1744,19 @@ static int
 check_throttles( connecttab* c )
     {
     int tnum;
+    long l;
 
     c->numtnums = 0;
-    c->limit = THROTTLE_NOLIMIT;
+    c->max_limit = c->min_limit = THROTTLE_NOLIMIT;
     for ( tnum = 0; tnum < numthrottles && c->numtnums < MAXTHROTTLENUMS;
 	  ++tnum )
 	if ( match( throttles[tnum].pattern, c->hc->expnfilename ) )
 	    {
 	    /* If we're way over the limit, don't even start. */
-	    if ( throttles[tnum].rate > throttles[tnum].limit * 2 )
+	    if ( throttles[tnum].rate > throttles[tnum].max_limit * 2 )
+		return 0;
+	    /* Also don't start if we're under the minimum. */
+	    if ( throttles[tnum].rate < throttles[tnum].min_limit )
 		return 0;
 	    if ( throttles[tnum].num_sending < 0 )
 		{
@@ -1590,8 +1765,16 @@ check_throttles( connecttab* c )
 		}
 	    c->tnums[c->numtnums++] = tnum;
 	    ++throttles[tnum].num_sending;
-	    c->limit = MIN(
-		c->limit, throttles[tnum].limit / throttles[tnum].num_sending );
+	    l = throttles[tnum].max_limit / throttles[tnum].num_sending;
+	    if ( c->max_limit == THROTTLE_NOLIMIT )
+		c->max_limit = l;
+	    else
+		c->max_limit = MIN( c->max_limit, l );
+	    l = throttles[tnum].min_limit;
+	    if ( c->min_limit == THROTTLE_NOLIMIT )
+		c->min_limit = l;
+	    else
+		c->min_limit = MAX( c->min_limit, l );
 	    }
     return 1;
     }
@@ -1623,35 +1806,37 @@ update_throttles( ClientData client_data, struct timeval* nowP )
 	      throttles[tnum].bytes_since_avg / THROTTLE_TIME ) / 4;
 	throttles[tnum].bytes_since_avg = 0;
 	/* Log a warning message if necessary. */
-	if ( throttles[tnum].rate > throttles[tnum].limit )
+	if ( throttles[tnum].rate > throttles[tnum].max_limit )
 	    {
-	    if ( throttles[tnum].rate > throttles[tnum].limit * 2 )
-		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld GREATLY exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].limit );
+	    if ( throttles[tnum].rate > throttles[tnum].max_limit * 2 )
+		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld GREATLY exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].max_limit );
 	    else
-		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].limit );
+		syslog( LOG_INFO, "throttle #%d '%.80s' rate %ld exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].max_limit );
+	    }
+	if ( throttles[tnum].rate < throttles[tnum].min_limit )
+	    {
+	    syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld lower than minimum %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].min_limit );
 	    }
 	}
     }
 
 
+static void
+finish_connection( connecttab* c, struct timeval* tvP )
+    {
+    /* If we haven't actually sent the buffered response yet, do so now. */
+    httpd_write_response( c->hc );
+
+    /* And clear. */
+    clear_connection( c, tvP );
+    }
+
+
 static void
 clear_connection( connecttab* c, struct timeval* tvP )
     {
     ClientData client_data;
 
-    /* If we haven't actually sent the buffered response yet, do so now. */
-    httpd_write_response( c->hc );
-
-    if ( c->idle_read_timer != (Timer*) 0 )
-	{
-	tmr_cancel( c->idle_read_timer );
-	c->idle_read_timer = 0;
-	}
-    if ( c->idle_send_timer != (Timer*) 0 )
-	{
-	tmr_cancel( c->idle_send_timer );
-	c->idle_send_timer = 0;
-	}
     if ( c->wakeup_timer != (Timer*) 0 )
 	{
 	tmr_cancel( c->wakeup_timer );
@@ -1669,16 +1854,24 @@ clear_connection( connecttab* c, struct timeval* tvP )
     ** circumstances that make a lingering close necessary.  If the flag
     ** isn't set we do the real close now.
     */
+    if ( c->conn_state == CNST_LINGERING )
+	{
+	/* If we were already lingering, shut down for real. */
+	tmr_cancel( c->linger_timer );
+	c->linger_timer = (Timer*) 0;
+	c->hc->should_linger = 0;
+	}
     if ( c->hc->should_linger )
 	{
 	c->conn_state = CNST_LINGERING;
 	fdwatch_del_fd( c->hc->conn_fd );
+	shutdown( c->hc->conn_fd, SHUT_WR );
 	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );
-	/* Make sure we are still in no-delay mode. */
-	httpd_set_ndelay( c->hc->conn_fd );
 	client_data.p = c;
+	if ( c->linger_timer != (Timer*) 0 )
+	    syslog( LOG_ERR, "replacing non-null linger_timer!" );
 	c->linger_timer = tmr_create(
-	    tvP, linger_clear_connection, client_data, LINGER_TIME * 1000L, 0 );
+	    tvP, linger_clear_connection, client_data, LINGER_TIME, 0 );
 	if ( c->linger_timer == (Timer*) 0 )
 	    {
 	    syslog( LOG_CRIT, "tmr_create(linger_clear_connection) failed" );
@@ -1708,36 +1901,38 @@ really_clear_connection( connecttab* c, struct timeval* tvP )
 
 
 static void
-idle_read_connection( ClientData client_data, struct timeval* nowP )
+idle( ClientData client_data, struct timeval* nowP )
     {
+    int cnum;
     connecttab* c;
 
-    c = (connecttab*) client_data.p;
-    c->idle_read_timer = (Timer*) 0;
-    if ( c->conn_state != CNST_FREE )
+    for ( cnum = 0; cnum < maxconnects; ++cnum )
 	{
-	syslog( LOG_INFO,
-	    "%.80s connection timed out reading",
-	    httpd_ntoa( &c->hc->client_addr ) );
-	httpd_send_err( c->hc, 408, httpd_err408title, "", httpd_err408form, "" );
-	clear_connection( c, nowP );
-	}
-    }
-
-
-static void
-idle_send_connection( ClientData client_data, struct timeval* nowP )
-    {
-    connecttab* c;
-
-    c = (connecttab*) client_data.p;
-    c->idle_send_timer = (Timer*) 0;
-    if ( c->conn_state != CNST_FREE )
-	{
-	syslog( LOG_INFO,
-	    "%.80s connection timed out sending",
-	    httpd_ntoa( &c->hc->client_addr ) );
-	clear_connection( c, nowP );
+	c = &connects[cnum];
+	switch ( c->conn_state )
+	    {
+	    case CNST_READING:
+	    if ( nowP->tv_sec - c->active_at >= IDLE_READ_TIMELIMIT )
+		{
+		syslog( LOG_INFO,
+		    "%.80s connection timed out reading",
+		    httpd_ntoa( &c->hc->client_addr ) );
+		httpd_send_err(
+		    c->hc, 408, httpd_err408title, "", httpd_err408form, "" );
+		finish_connection( c, nowP );
+		}
+	    break;
+	    case CNST_SENDING:
+	    case CNST_PAUSING:
+	    if ( nowP->tv_sec - c->active_at >= IDLE_SEND_TIMELIMIT )
+		{
+		syslog( LOG_INFO,
+		    "%.80s connection timed out sending",
+		    httpd_ntoa( &c->hc->client_addr ) );
+		clear_connection( c, nowP );
+		}
+	    break;
+	    }
 	}
     }
 
@@ -1772,6 +1967,7 @@ occasional( ClientData client_data, struct timeval* nowP )
     {
     mmc_cleanup( nowP );
     tmr_cleanup();
+    watchdog_flag = 1;		/* let the watchdog know that we are alive */
     }
 
 
@@ -1803,7 +1999,7 @@ logstats( struct timeval* nowP )
     if ( stats_secs == 0 )
 	stats_secs = 1;	/* fudge */
     stats_time = now;
-    syslog( LOG_NOTICE,
+    syslog( LOG_INFO,
 	"up %ld seconds, stats for %ld seconds:", up_secs, stats_secs );
 
     thttpd_logstats( stats_secs );
@@ -1818,11 +2014,21 @@ logstats( struct timeval* nowP )
 static void
 thttpd_logstats( long secs )
     {
-    syslog( LOG_NOTICE,
-	"  thttpd - %ld connections (%g/sec), %d max simultaneous, %ld bytes (%g/sec), %d httpd_conns allocated",
-	stats_connections, (float) stats_connections / secs,
-	stats_simultaneous, stats_bytes, (float) stats_bytes / secs,
-	httpd_conn_count );
-    stats_connections = stats_bytes = 0L;
+    if ( secs > 0 )
+#ifdef HAVE_INT64T
+	syslog( LOG_INFO,
+	    "  thttpd - %ld connections (%g/sec), %d max simultaneous, %lld bytes (%g/sec), %d httpd_conns allocated",
+	    stats_connections, (float) stats_connections / secs,
+	    stats_simultaneous, (int64_t) stats_bytes,
+	    (float) stats_bytes / secs, httpd_conn_count );
+#else /* HAVE_INT64T */
+	syslog( LOG_INFO,
+	    "  thttpd - %ld connections (%g/sec), %d max simultaneous, %ld bytes (%g/sec), %d httpd_conns allocated",
+	    stats_connections, (float) stats_connections / secs,
+	    stats_simultaneous, (long) stats_bytes,
+	    (float) stats_bytes / secs, httpd_conn_count );
+#endif /* HAVE_INT64T */
+    stats_connections = 0;
+    stats_bytes = 0;
     stats_simultaneous = 0;
     }
